package org.netkernel.text.search.endpoint;

import java.io.File;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.net.URI;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.document.Field.Index;
import org.apache.lucene.document.Field.Store;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.queryParser.MultiFieldQueryParser;
import org.apache.lucene.search.BooleanClause;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.Sort;
import org.apache.lucene.search.TopDocs;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.FSDirectory;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.text.search.representation.LuceneConfigRep;

public class LuceneAccessor extends StandardAccessorImpl
{
	public static final int DEFAULT_MAX_FIELD_LENGTH = 25000;

	private static final String GOLDEN_THREAD_PREFIX = "gt:searchIndex:";

	public static final String SEARCH_CONFIG_RESOURCE = "res:/etc/SearchConfig.xml";

	public static final String DEFAULT_SEARCH_CONFIG_RESOURCE = "res:/etc/DefaultSearchConfig.xml";

	public LuceneAccessor()
	{
		this.declareThreadSafe();
	}

	@Override
	public void onSource(INKFRequestContext context) throws Exception
	{	LuceneConfigRep config = null;
		String configURI = null;
		if (context.getThisRequest().argumentExists("config"))
		{	configURI = context.getThisRequest().getArgumentValue("config");
		} else
		{	if (context.exists(SEARCH_CONFIG_RESOURCE))
			{	configURI = SEARCH_CONFIG_RESOURCE;
			} else
			{	configURI = DEFAULT_SEARCH_CONFIG_RESOURCE;
			}
		}
		config = context.source(configURI, LuceneConfigRep.class);

		String directory = null;
		if (context.getThisRequest().argumentExists("directory"))
		{	directory = context.getThisRequest().getArgumentValue("directory");
		} else
		{	directory = config.getDirectory();
		}
		if (directory == null)
		{	throw new Exception(
					"index directory configuration is not present as argument or in configuration");
		}

		String type = context.getThisRequest().getArgumentValue("activeType");
		if (type.equals("textSearch"))
		{	Query q = context.source("arg:operand", Query.class);
			long timer=start();
			IndexSearcher isearcher = new IndexSearcher(getDirectory(directory));
			int limit=100;
			if(context.getThisRequest().argumentExists("limit"))
			{  try
				{	limit=Integer.parseInt(context.getThisRequest().getArgumentValue("limit"));
				}
				catch(Exception e)
				{	context.logRaw(INKFRequestContext.LEVEL_WARNING, "Invalid limit value for search - defaulting to limit 100");					
				}
			}
			TopDocs tdocs = isearcher.search(q, null, limit, Sort.RELEVANCE);
			double time=stop(timer);
			HDSBuilder b = new HDSBuilder();
			b.pushNode("hits");
			b.addNode("time", time);
			b.addNode("total", tdocs.scoreDocs.length);
			for (int i = 0; i < tdocs.scoreDocs.length; i++)
			{	ScoreDoc sd = tdocs.scoreDocs[i];
				b.pushNode("hit");
				b.addNode("score", sd.score);
				Document doc = isearcher.doc(sd.doc);
				List<Field> docfields = doc.getFields();
				for (Iterator<Field> it = docfields.iterator(); it.hasNext();)
				{	Field f = it.next();
					b.pushNode("field");
					b.addNode("name", f.name());
					b.addNode("value", f.stringValue());
					b.popNode();
				}
				b.popNode();
			}
			b.popNode();
			INKFRequest req = context.createRequest("active:attachGoldenThread");
			req.addArgument("id", GOLDEN_THREAD_PREFIX + directory);
			context.issueRequestForResponse(req);
			context.createResponseFrom(b.getRoot());
		} else if (type.equals("textIndexDelete"))
		{	File f = getDirectoryFile(directory);
			INKFResponse resp = context.createResponseFrom(recurseDelete(f));
			resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
			cutGoldenThread(context, directory);
		} else if (type.equals("textIndex"))
		{	IndexWriter iwriter = getIndexWriter(directory, config);
			IHDSNode opd = context.source("arg:operand", IHDSNode.class);
			IHDSNode item = opd.getFirstNode("item");
			long timer=start();
			index(context, iwriter, item);
			iwriter.commit();
			iwriter.close();
			double time=stop(timer);
			setIndexResponse(context, 1, time);
			cutGoldenThread(context, directory);
		} else if (type.equals("textIndexBatch"))
		{	IndexWriter iwriter = getIndexWriter(directory, config);
			IHDSNode opd = context.source("arg:operand", IHDSNode.class);
			IHDSNodeList nl = opd.getNodes("/batch/item");
			long timer=start();
			for (int i = 0; i < nl.size(); i++)
			{	IHDSNode n = nl.get(i);
				index(context, iwriter, n);
			}
			iwriter.commit();
			iwriter.close();
			double time=stop(timer);
			setIndexResponse(context, nl.size(), time);
			cutGoldenThread(context, directory);
		}
	}
	
	private void setIndexResponse(INKFRequestContext context, int count, double time)
	{	HDSBuilder b=new HDSBuilder();
		b.pushNode("indexResult");
			b.addNode("time",time);
			b.addNode("count", count);
		b.popNode();
		INKFResponse resp=context.createResponseFrom(b.getRoot());
		resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);			
	}

	private void cutGoldenThread(INKFRequestContext context, String id)
			throws Exception
	{	INKFRequest req = context.createRequest("active:cutGoldenThread");
		req.addArgument("id", GOLDEN_THREAD_PREFIX + id);
		context.issueRequest(req);
	}

	private void index(INKFRequestContext context, IndexWriter iw, IHDSNode n) throws Exception
	{	Document doc=new Document();
		IHDSNodeList nl=n.getNodes("field");
		for(int i=0; i<nl.size(); i++)
		{	IHDSNode field=nl.get(i);
			String name=(String)field.getFirstValue("name");
			String store,index,value=null;
			boolean isURI=false;
			IHDSNode part=field.getFirstNode("@store");
			if(part!=null)
			{	store=(String)part.getValue();					
			}
			else
			{	store="NO";				
			}
			part=field.getFirstNode("@index");
			if(part!=null)
			{	index=(String)part.getValue();				
			}
			else
			{	index="ANALYZED";			
			}
			part=field.getFirstNode("uri");
			if(part!=null)
			{	isURI=true;
				value=(String)part.getValue();
			}
			else
			{	part=field.getFirstNode("value");
				if(part!=null)
				{	value=(String)part.getValue();
				}
			}
			if(value==null)
			{	throw new Exception("uri or value is required on field");				
			}
			Field.Index idx=(Field.Index)Field.Index.class.getField(index).get(null);
			Field.Store str=(Field.Store)Field.Store.class.getField(store).get(null);
			if(isURI)
			{	IReadableBinaryStreamRepresentation rbs=(IReadableBinaryStreamRepresentation)context.source(value, IReadableBinaryStreamRepresentation.class);
				if(rbs!=null)
				{
					doc.add(new Field(name, new InputStreamReader(rbs.getInputStream())));
					doc.add(new Field(name+"URI", value, Store.YES, Index.NO));
				}
				else
				{	context.logRaw(INKFRequestContext.LEVEL_WARNING, "Cannot index resource ["+value+"]  as it is null");					
				}
			}
			else
			{	doc.add(new Field(name, value, str, idx));
			}
		}
		iw.addDocument(doc);
	}

	private IndexWriter getIndexWriter(String directory, LuceneConfigRep config)
			throws Exception
	{
		Analyzer a = config.getAnalyzer();
		Directory d = getDirectory(directory);
		return new IndexWriter(d, a, new IndexWriter.MaxFieldLength(config
				.getMaxFieldLength()));
	}

	private Directory getDirectory(String dir) throws Exception
	{	return FSDirectory.getDirectory(getDirectoryFile(dir));
	}

	private File getDirectoryFile(String dir) throws Exception
	{	File dirf = new File(URI.create(dir));
		if (!dirf.exists())
		{
			dirf.mkdirs();
		}
		if (!dirf.isDirectory())
		{
			throw new Exception(
					"'directory' argument must be a file: URI to a directory. ");
		}
		return dirf;
	}
	
	private long start()
	{	return System.currentTimeMillis();		
	}

	private double stop(long startTime)
	{	return (System.currentTimeMillis()-startTime)/1000.0;		
	}
	
	private boolean recurseDelete(File f)
	{	if (f.isDirectory())
		{
			File[] l = f.listFiles();
			for (int i = 0; i < l.length; i++)
			{
				recurseDelete(l[i]);
			}
		}
		return f.delete();
	}
}
